/*
 * cref - Count cross references between block groups
 *
 * Given an ext2/3 file system image, figure out how many
 * cross-block-group references there are.  Example: a file with
 * blocks in two groups creates one cross reference, a file with
 * blocks in one group has zero cross references.
 *
 * (C) 2006-2007 Karuna Sagar K <karunasagark@gmail.com>
 * (C) 2007 Val Henson <val@nmt.edu>
 *
 * Licensed under GPLv2
 */

#include <unistd.h>
#include <stdio.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <error.h>
#include <errno.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>
#include <math.h>
#include <sys/param.h>

#include <linux/fs.h>

/*
 * Various counters
 */

static unsigned long long all_dirdir_crefs;
static unsigned long long all_dirfile_crefs;
static unsigned long long all_file_crefs;
static unsigned long long all_dir_crefs;
static unsigned long long all_crefs;
static unsigned long long dir_count;
static unsigned long long file_count;
static unsigned long long other_count;
/* Count of cross references between every pair of groups */
static unsigned long long **group_crefs;

/*
 * Command line arguments
 */

static unsigned long long blk_per_grp;
static unsigned long long free_blk_count;
static unsigned long long blk_grp_count;
static unsigned long long inodes_per_grp;
static char *mountpoint;
static unsigned long long blk_count;
static unsigned long long blk_sz;
static unsigned int verbose;

/*
 * Random global variables
 */
static char *cmd;

static void
usage(void)
{
	fprintf(stderr, "Usage: %s [options]\n"
		"\t -b\t <blocks per group>\n"
		"\t -f\t <# free blocks>\n"
		"\t -g\t <# block groups>\n"
		"\t -i\t <# inodes per groups>\n"
		"\t -m\t <mountpoint>\n"
		"\t -n\t <# blocks>\n"
		"\t -s\t <block size>\n"
		"\t -v\t verbose\n",
		cmd);
	exit(1);
}

static unsigned long long
count_file_references(char *f, unsigned long long inode)
{
	unsigned long long fileblks, blk, curblk;
	unsigned long long crefs = 0;
	unsigned long long inode_grp, blk_grp;
	unsigned long long found_in_grp[blk_grp_count];
	unsigned int i;
	off_t filesz;
	int fd;

	for (i = 0; i < blk_grp_count; i++)
		found_in_grp[i] = 0;

	file_count++;

	if (verbose)
		printf("F: %s (%llu): ", f, inode);

	inode_grp = (inode - 1) / inodes_per_grp;
	found_in_grp[inode_grp]++;

	if (verbose)
		printf("%llu, ", inode_grp);
	
	if ((fd = open(f, O_RDONLY | O_NONBLOCK)) == -1)
		error(1, errno, "Error opening file %s", f);

	filesz = lseek(fd, 0, SEEK_END);
	fileblks = (filesz + blk_sz - 1) / blk_sz;
	lseek(fd, 0, SEEK_SET);

	for (blk = 0; blk < fileblks; blk++) {
		curblk = blk;
		if (ioctl(fd, FIBMAP, &curblk) == -1) {
			/* Usually this is harmless - special file, etc. */
			if (verbose)
				error(0, errno,"%s: block %llu: ioctl failed",
				      f, blk);
		}
		blk_grp = curblk / blk_per_grp;
		found_in_grp[blk_grp]++;
		if (found_in_grp[blk_grp] == 1) {
			crefs++;
			group_crefs[inode_grp][blk_grp]++;
			group_crefs[blk_grp][inode_grp]++;
			if (verbose)
				printf("%llu, ", blk_grp);
		}
	}
	if (verbose)
		printf("= %llu\n", crefs);

	all_file_crefs += crefs;
	close(fd);
	return crefs;
}

static int
count_dir_references(char *dir, unsigned long long dir_inode)
{
	DIR *d;
	struct dirent *dentry;
	struct stat stbuf;
	char name[MAXPATHLEN];
	/* Directory-directory cross references in this dir */
	unsigned long long dirdir_crefs = 0;
	/* Directory-file cross references in this dir */
	unsigned long long dirfile_crefs = 0;
	/* Directory-special file cross references in this dir */
	unsigned long long dirspecial_crefs = 0;
	/* Cross references between this directory and its entries */
	unsigned long long crefs = 0;
	/* All cross references of tree below this directory */
	unsigned long long subdir_crefs = 0;
	unsigned long long inode;
	unsigned long long dir_grp, dentry_grp;
	unsigned long long found_in_grp[blk_grp_count];
	unsigned int i;

	for (i = 0; i < blk_grp_count; i++)
		found_in_grp[i] = 0;

	dir_count++;

	if (verbose)
		printf("D: %s (%lu): ", dir, stbuf.st_ino);

	dir_grp = (dir_inode - 1) / inodes_per_grp;
	found_in_grp[dir_grp]++;

	if (verbose)
		printf("%llu, ", dir_grp);
	
	if ((d = opendir(dir)) == NULL)
		error(1, errno, "Error opening directory: %s\n", dir);

	while((dentry = readdir(d)) != NULL) {
		if (strcmp(dentry->d_name, ".") == 0 ||
		    strcmp(dentry->d_name, "..") == 0)
			continue;

		sprintf(name, "%s/%s", dir, dentry->d_name);

		if (lstat(name, &stbuf))
			error(1, errno, "Error stating dir/file: %s\n", name);
		inode = stbuf.st_ino;

		dentry_grp = (inode - 1) / inodes_per_grp;
		found_in_grp[dentry_grp]++;

		/* Count the cross references by file type */
		if (found_in_grp[dentry_grp] == 1) {
			group_crefs[dir_grp][dentry_grp]++;
			group_crefs[dentry_grp][dir_grp]++;
			if (S_ISDIR(stbuf.st_mode))
				dirdir_crefs++;
			else if (S_ISREG(stbuf.st_mode))
				dirfile_crefs++;
			else /* Links, block, char, pipes, etc. */
				dirspecial_crefs++;
			if (verbose)
				printf("%llu, ", dentry_grp);
		}

		/* Now count cross references in the subtree below this dir */
		if (S_ISDIR(stbuf.st_mode))
			subdir_crefs += count_dir_references(name, inode);
		else if (S_ISREG(stbuf.st_mode))
		        subdir_crefs += count_file_references(name, inode);
		else
			other_count++;
	}

	crefs = dirfile_crefs + dirdir_crefs + dirspecial_crefs;
	if (verbose)
		printf("= %llu\n", crefs);

	all_dirdir_crefs += dirdir_crefs;
	all_dirfile_crefs += dirfile_crefs + dirspecial_crefs;

	closedir(d);
	return crefs + subdir_crefs;
}

#define SEEK_TIME		0.008	/* 8 milliseconds */
#define ROTATIONAL_LATENCY	0.0074	/* 7.4 milliseconds */
#define CHECK_CREF_TIME		(SEEK_TIME + ROTATIONAL_LATENCY)

#define SEEK_TIME_2013		0.0065	/* 6.5 milliseconds */
#define ROTATIONAL_LATENCY_2013	0.0047	/* 4.7 milliseconds */
#define CHECK_CREF_TIME_2013	(SEEK_TIME_2013 + ROTATIONAL_LATENCY_2013)

static void
process_group_stats(void)
{
	unsigned long long totals[blk_grp_count];
	unsigned long long max = 0;
	unsigned long long max_grp;
	double average = 0;
	unsigned long long i, j;

	for (i = 0; i < blk_grp_count; i++) {
		totals[i] = 0;
		for (j = 0; j < blk_grp_count; j++)
			totals[i] += group_crefs[i][j];
		if (totals[i] > max) {
			max = totals[i];
			max_grp = i;
		}
	}
	average = (double) all_crefs / (double) blk_grp_count;
	printf("Average cross references per group = %llu\n",
	       (unsigned long long) average);
	printf("Maximum cross references in a group = %llu\n", max);
	printf("Max group is %llu (", max_grp);
	for (i = 0; i < blk_grp_count; i++)
		if (group_crefs[max_grp][i] != 0)
			printf("%llu:%llu, ", i, group_crefs[max_grp][i]);

	printf(")\n");
	printf("Average additional time to check cross references = %.2f s\n",
	       average * CHECK_CREF_TIME);
	printf("Max additional time to check cross references = %.2f s\n",
	       max * CHECK_CREF_TIME);
	printf("2013 average additional time to check cross references = %.2f s\n",
	       average * CHECK_CREF_TIME_2013);
	printf("2013 max additional time to check cross references = %.2f s\n",
	       max * CHECK_CREF_TIME_2013);
}

int
main(int argc, char *argv[])
{
	struct stat stbuf;
	unsigned long long total = 0;
	unsigned long long inode;
	unsigned int i;
	char c;

	cmd = argv[0];

	while ((c = getopt(argc, argv, "b:f:g:i:m:n:s:v")) != -1) {
		switch (c) {
		case 'b':
			blk_per_grp = atoi(optarg);
			if (blk_per_grp == 0)
				usage();
			break;
		case 'f':
			free_blk_count = atoi(optarg);
			if (free_blk_count == 0)
				usage();
			break;
		case 'g':
			blk_grp_count= atoi(optarg);
			if (blk_grp_count == 0)
				usage();
			break;
		case 'i':
			inodes_per_grp = atoi(optarg);
			if (inodes_per_grp == 0)
				usage();
			break;
		case 'm':
			mountpoint = strdup(optarg);
			if (mountpoint == NULL) {
			        perror("Couldn't allocate mountpoint string\n");
				exit(1);
			}
			break;
		case 'n':
			blk_count = atoi(optarg);
			if (blk_count == 0)
				usage();
			break;
		case 's':
			blk_sz = atoi(optarg);
			if (blk_sz == 0)
				usage();
			break;
		case 'v':
			verbose++;
			break;
		default:
			usage();
		}
	}

	group_crefs = malloc(sizeof (unsigned long long) * blk_grp_count);
	if (group_crefs == NULL)
		error(1, ENOMEM, NULL);
	for (i = 0; i < blk_grp_count; i++) {
		group_crefs[i] = malloc(sizeof (unsigned long long) *
					blk_grp_count);
		if (group_crefs[i] == NULL)
			error(1, ENOMEM, NULL);
		memset(group_crefs[i], 0,
		       sizeof (unsigned long long) * blk_grp_count);
	}

	if (lstat(mountpoint, &stbuf))
		error(1, errno, "Error stating dir: %s\n", mountpoint);
	inode = stbuf.st_ino;

	total = count_dir_references(mountpoint, inode);

	all_dir_crefs = all_dirdir_crefs + all_dirfile_crefs;
	all_crefs = all_dir_crefs + all_file_crefs;

	printf("Total size = %llu KB\n", blk_count*blk_sz/1024);
	printf("Total data stored = %llu KB\n", blk_sz*(blk_count-free_blk_count)/1024);
	printf("Number of files = %llu\n", file_count);
	printf("Number of directories = %llu\n", dir_count);
	printf("Number of special files = %llu\n", other_count);
	printf("Size of block groups = %llu KB\n", blk_sz*blk_per_grp/1024);
	printf("Inodes per block group = %llu\n", inodes_per_grp);
	printf("Intra-file cross references = %llu\n", all_file_crefs);
	printf("Directory-subdirectory references = %llu\n", all_dirdir_crefs);
	printf("Directory-file references = %llu\n", all_dirfile_crefs);
	printf("Total directory cross references = %llu\n", all_dir_crefs);
	printf("Total cross references = %llu\n", all_dir_crefs +
	       all_file_crefs);
	printf("Total cross references = %llu\n", total);

	process_group_stats();

	return 0;
}
